import functools
import typing as t

T_Callable = t.TypeVar("T_Callable", bound=t.Callable[..., t.Any])
depth = 0


def debug_trace(func: T_Callable) -> T_Callable:
    return func

    @functools.wraps(func)  # type: ignore[unreachable]
    def wrapper(*a, **kw):
        global depth
        print(depth * "| " + "enter", func.__name__)
        try:
            depth += 1
            try:
                result = func(*a, **kw)
            finally:
                depth -= 1
        except Exception as exc:
            print(depth * "| " + "error in", func.__name__, exc)
            raise
        else:
            print(depth * "| " + "return", func.__name__, repr(result))
            return result

    return wrapper



"""
Transpile BuildDSL code to full fledged Python code.
"""

import ast
import logging
import sys
import typing as t
from dataclasses import dataclass, field

from ._rewriter import Closure, Grammar, Rewriter
from .ast_utils import DynamicLookupRewriter


@dataclass
class TranspileOptions:
    """Options for transpiling BuildDSL code."""

    #: If enabled, names are read, written and deleted through the `__getitem__()`, `__setitem__()` and
    #: `__delitem__()` of the given name. If you need extra flexibility, you can set this to the name of
    #: a global object that is then responsible for name resolution.
    closure_target: t.Optional[str] = None  # '__closure__'

    #: Set of builtin names that are "pure", i.e. they are never touched by the #NameRerwriter.
    #: This is only used if #closure_target is set.
    pure_builtins: t.Collection[str] = frozenset()  # frozenset(['__closure_decorator__'])

    #: This is only used if #closure_target is specified and the #NameRewriter kicks in. Variable declarations
    #: prefixed with `def` are prefixed with the given string. Defaults to prefix supplied in the #grammar.
    local_vardef_prefix: str = "_def_"

    #: A preamble of pure Python code to include at the top of the module.
    #: Example: `'from myruntime import __closure_decorator__\n'`
    preamble: str = ""

    #: Pure python code to include before a closure definition, for example to decorate it.
    #: Example: `'@__closure_decorator__(__closure__)\n'`
    closure_def_prefix: str = ""

    #: The default argument list for closures without an explicit argument list. By default a
    #: closure always accepts a "self" argument, but also any other arguments that are passed
    #: to it. This is useful when using a arglist-less closure for a function that passed
    #: multiple arguments, but the closure is only interested in the first.
    closure_default_arglist: str = "self, *arguments, **kwarguments"

    #: A prefix to the argument list of closures. This prefix is added even to closures without
    #: explicit arglist (so the final arglist will be #closure_arglist_prefix followed by
    #: #closure_default_arglist).
    closure_arglist_prefix: str = ""  # '__closure__,'

    grammar: Grammar = field(default_factory=Grammar)

    def sync(self) -> None:
        """Synchronize the options to the #Grammar settings, i.e. the #Grammar.local_def setting
        will be enabled or disabled depending on whether #closure_target is set and the
        #local_vardef_prefix is used to set #Grammar.local_prefix."""

        self.grammar.local_def = self.closure_target is not None
        self.grammar.local_prefix = self.local_vardef_prefix


def transpile_to_ast(code: str, filename: str, options: t.Optional[TranspileOptions] = None) -> ast.Module:
    """
    Transpile the BuildDSL *code* to a Python `ast.Module` that can be executed.
    """

    options = options or TranspileOptions()
    rewrite = Rewriter(code, filename, options.grammar).rewrite()
    if sys.version_info[:2] <= (3, 7):
        module = ast.parse(rewrite.code, filename, mode="exec")
    else:
        module = ast.parse(rewrite.code, filename, mode="exec", type_comments=False)
    module = ClosureRewriter(filename, options, rewrite.closures).visit(module)
    if options.closure_target:
        dynamic_lookup = DynamicLookupRewriter(
            options.closure_target, options.pure_builtins, options.local_vardef_prefix
        )
        module = t.cast(ast.Module, dynamic_lookup.visit(module))
    return ast.fix_missing_locations(module)


def transpile_to_source(code: str, filename: str, options: t.Optional[TranspileOptions] = None) -> str:
    """
    Transpile the BuildDSL *code* to Python code. Requires the `astor` module to be installed.
    """

    from astor import to_source  # type: ignore

    return to_source(transpile_to_ast(code, filename, options))  # type: ignore[no-any-return]


class ClosureRewriter(ast.NodeTransformer):
    """
    Rewrites references to closure variables and injects Closure function definitions.
    """

    log = logging.getLogger(__module__ + "." + __qualname__)  # type: ignore

    def __init__(self, filename: str, options: TranspileOptions, closures: t.Dict[str, Closure]) -> None:
        self.filename = filename
        self.options = options
        self.closures = closures

        # Keep track of the hierarchy during the visitation.
        self._hierarchy: t.List[ast.AST] = []

        # Marks the statement nodes in the hierarchy with the closure name(s) to insert.
        self._closure_inserts: t.Dict[ast.stmt, t.List[str]] = {}

    def _get_closure_def(self, closure_id: str) -> ast.FunctionDef:
        """
        Generate a function definition for a closure id.
        """

        closure = self.closures[closure_id]
        if closure.parameters is None:
            arglist = self.options.closure_default_arglist
        else:
            arglist = ", ".join(closure.parameters)
        arglist = self.options.closure_arglist_prefix + arglist

        function_code = f"{self.options.closure_def_prefix}def {closure_id}({arglist}):\n"
        function_code = "\n" * (function_code.count("\n") + closure.line) + function_code
        if closure.expr:
            function_code += " " * closure.indent + "return " + closure.expr
        else:
            function_code += (closure.body or "").rstrip() or (" " * closure.indent + "pass")

        # self.log.debug('_get_closure_def(%r): parse function body\n\n%s\n', closure_id,
        #                '  ' + '\n  '.join(function_code.lstrip().splitlines()))

        if sys.version_info[:2] <= (3, 7):
            module = ast.parse(function_code, self.filename, mode="exec")
        else:
            module = ast.parse(function_code, self.filename, mode="exec", type_comments=False)

        func = module.body[0]
        assert isinstance(func, ast.FunctionDef)
        return func

    def visit_Name(self, name: ast.Name) -> ast.AST:
        if name.id in self.closures:
            for node in reversed(self._hierarchy):
                if isinstance(node, ast.stmt):
                    self._closure_inserts.setdefault(node, []).append(name.id)
                    break
                elif isinstance(node, (ast.FunctionDef, ast.ClassDef)):
                    raise RuntimeError("did not find inner statement to inject closure")
        return self.generic_visit(name)

    def visit_Module(self, node: ast.Module) -> ast.AST:
        preamble = ast.parse(self.options.preamble, self.filename, mode="exec")
        node.body[0:0] = preamble.body
        return self.generic_visit(node)

    def visit(self, node: ast.AST) -> t.Any:
        self._hierarchy.append(node)
        try:
            result: t.Union[ast.AST, t.List[ast.AST]] = super().visit(node)
            if node in self._closure_inserts:
                assert isinstance(node, ast.stmt)
                assert isinstance(result, ast.AST)
                result = [result]
                for closure_id in self._closure_inserts.get(node, []):
                    func = self.visit(self._get_closure_def(closure_id))
                    result.insert(len(result) - 1, func)
            return result
        finally:
            assert self._hierarchy.pop() == node



import typing as t

from ._transpiler import TranspileOptions, transpile_to_ast


def execute(
    code: t.Union[str, t.TextIO],
    filename: t.Optional[str],
    globals: t.Dict[str, t.Any],
    locals: t.Optional[t.Mapping[str, t.Any]] = None,
    options: t.Optional[TranspileOptions] = None,
) -> None:
    """
    Executes BuildDSL code in the context specified with *globals* and *locals*.

    @param code: The code to execute.
    @param filename: The filename where the code is from; shown in errors.
    @param globals: The globals for the code.
    @param locals: The locals for the code.
    @param options: Options for the DSL transpiler.
    """

    if hasattr(code, "read"):
        code = t.cast(t.TextIO, code).read()
        filename = getattr(code, "name", None)

    assert isinstance(code, str)
    filename = filename or "<string>"

    ast = transpile_to_ast(code, filename, options)
    compiled_code = compile(ast, filename, "exec")
    exec(compiled_code, globals, locals or globals)



import argparse
import importlib
import os
import sys

from . import Closure, execute, transpile_to_source

parser = argparse.ArgumentParser(prog=os.path.basename(sys.executable) + " -m builddsl")
parser.add_argument("file", nargs="?")
parser.add_argument("-c", "--context", metavar="ENTRYPOINT")
parser.add_argument("-E", "--transpile", action="store_true")
parser.add_argument("-C", "--enable-closures", action="store_true")


def main() -> None:
    args = parser.parse_args()

    if args.transpile:
        if args.context:
            parser.error("conflicting arguments: -c/--context and -E/--transpile")

    if args.file:
        with open(args.file) as fp:
            code = fp.read()
        filename = args.file
    else:
        code = sys.stdin.read()
        filename = "<stdin>"

    options = Closure.get_options() if args.enable_closures else None

    if args.transpile:
        print(transpile_to_source(code, filename, options))
        return

    if args.context:
        module_name, member = args.context.partition(":")
        context = getattr(importlib.import_module(module_name), member)()
    else:
        context = None

    globals_ = {"self": context} if context is not None else {}
    execute(code, filename, globals_, None, options)


if __name__ == "__main__":
    main()



"""
Rewrite BuildDSL code to pure Python code.
"""

import contextlib
import enum
import logging
import re
import string
import sys
import typing as t
from dataclasses import dataclass

from nr.io.lexer import Cursor, ProxyToken as _ProxyToken, RuleSet, Tokenizer, rules

try:
    from termcolor import colored
except ImportError:

    def colored(s, *a, **kw) -> str:  # type: ignore
        return str(s)


from ._util import debug_trace

logger = logging.getLogger(__name__)


@dataclass
class Grammar:
    """
    Grammar settings.
    """

    #: Whether unparenthesized function calls are allowed.
    unparen_calls: bool = True

    #: Whether keyword arguments may be specified using colons (`:`) as well as equal signs (`=`).
    colon_kwargs: bool = True

    #: Accept function arguments with comma separation.
    nocomma_args: bool = True

    #: Whether the `def varname = ...` syntax is allowed and understood. This is an important
    #: syntax feature when enabling the #NameRewriter with #TranspileOptions.closure_target.
    local_def: bool = False

    #: The keyword for local definitions.
    local_keyword: str = "def"

    #: The prefix for variable names that are localized. This is picked up by the transpiler.
    local_prefix: str = "_def_"


class Token(enum.Enum):
    Eof = enum.auto()
    Indent = enum.auto()
    Whitespace = enum.auto()
    Newline = enum.auto()
    Comment = enum.auto()
    Name = enum.auto()
    Literal = enum.auto()
    Control = enum.auto()


PYTHON_BLOCK_KEYWORDS = frozenset(["class", "def", "if", "elif", "else", "for", "while", "with"])
ASSIGNMENT_OPERATORS = ["=", "+=", "-=", "*=", "/=", "%=", "//=", "**=", "&=", "|=", "^=", ">>=", "<<=", "@="]
BINARY_OPERATORS = [x[:-1] for x in ASSIGNMENT_OPERATORS[1:]] + [
    ".",
    "<",
    ">",
    "==",
    "<=",
    ">=",
    "!=",
    ":=",
    "is",
    "and",
    "or",
]
UNARY_OPERATORS = ["not", "~"]
OTHER_CONTROL_CHARACTERS = list("()[]{},:;") + ["->"]
_ALL_CONTROL_CHARACTERS = sorted(
    ASSIGNMENT_OPERATORS + BINARY_OPERATORS + UNARY_OPERATORS + OTHER_CONTROL_CHARACTERS,
    key=len,
    reverse=True,
)
_WORD_CONTROL_CHARACTERS = [op for op in _ALL_CONTROL_CHARACTERS if op.isalpha()]

rule_set = RuleSet((Token.Eof, ""))
rule_set.rule(Token.Indent, rules.regex_extract(r"[\t ]*", at_line_start_only=True))
rule_set.rule(Token.Newline, rules.regex_extract(r"\n"))
rule_set.rule(Token.Whitespace, rules.regex_extract(r"\s+"))
rule_set.rule(Token.Comment, rules.regex_extract(r"#.*"))
rule_set.rule(Token.Control, rules.regex_extract("(" + "|".join(map(re.escape, _WORD_CONTROL_CHARACTERS)) + r")\b"))
rule_set.rule(Token.Name, rules.regex_extract(r"[A-Za-z\_][A-Za-z0-9\_]*"))
rule_set.rule(Token.Literal, rules.regex_extract(r"[+\-]?(\d+)(\.\d*)?"))
rule_set.rule(Token.Literal, rules.string_literal())
rule_set.rule(Token.Control, rules.regex_extract("|".join(map(re.escape, _ALL_CONTROL_CHARACTERS))))


class ProxyToken(_ProxyToken[Token, str]):
    """
    Extension class that adds some useful utility methods to test the contents of the token.
    """

    def is_ignorable(self, newlines: bool = False) -> bool:
        if newlines and self.type == Token.Newline:
            return True
        return self.type in (Token.Indent, Token.Whitespace, Token.Comment)

    def is_control(self, charpool: t.Collection[str]) -> bool:
        return self.type == Token.Control and self.value in charpool


class ParseMode(enum.IntFlag):
    """Flags that describe the current parse environment."""

    #: Nothing specific.
    DEFAULT = 0

    #: The currently parsed expression is grouped in parenthesis and may wrap over lines.
    GROUPED = 1 << 0

    #: The currently parsed expression is the outter parenthesis of a function call.
    FUNCTION_CALL = 1 << 1

    #: The currently parsed expression is an argument to a function call.
    CALL_ARGS = 1 << 2


@dataclass
class SyntaxError(Exception):
    """
    Specialized BuildDSL syntax error (the internal SyntaxError class is weird).

    If the `termcolor` module is installed, the error message will be color coded.
    """

    message: str
    filename: str
    line: int
    column: int
    text: str

    def get_text_hint(self) -> str:
        return "\n".join((self.text, "~" * self.column + "^"))

    def __str__(self) -> str:
        lines = [
            "",
            f'  in {colored(self.filename, "blue")} at line {self.line}: {colored(self.message, "red")}',
            *("  |" + line for line in self.get_text_hint().splitlines()),
        ]
        return "\n".join(lines)


@dataclass
class Closure:
    """
    Contains the definition of a closure in text format.
    """

    #: A unique ID for the closure, usually derived from the number of closures that have already
    #: been parsed in the same file or it's parent closures.
    id: str

    #: The line number where the closure begins.
    line: int

    #: The indentation of the closure's body. For a single expression, this represents the
    #: offset of the expression in line.
    indent: int

    #: The parameter names of the closure. May be `None` to indicate that closure had no header.
    parameters: t.Optional[t.List[str]]

    #: The body of the closure. May be `None` if the closure body is not constructed using curly
    #: braces to encapsulate multiple statements. In that case, the #expr field is set instead.
    body: t.Optional[str]

    #: Only set if the body of the closure is just an expression.
    expr: t.Optional[str]


@dataclass
class RewriteResult:
    """
    The result of rewriting BuildDSL code to pure Python code.
    """

    #: The BuildDSL code rewritten as Python code.
    code: str

    #: The closures extracted from the code.
    closures: t.Dict[str, Closure]


class Rewriter:
    """
    This class rewrites BuildDSL code to pure Python code. Closures are extracted from the code
    and replaced with whitespace where appropriate to keep line and column numbers of the code in
    tact as much as possible (not always fully accurate).
    """

    def __init__(self, text: str, filename: str, grammar: t.Optional[Grammar] = None) -> None:
        """
        # Arguments
        text: The BuildDSL code to parse and turn into an AST-like structure.
        filename: The filename where the DSL code is from.
        """

        self.tokenizer = Tokenizer(rule_set, text)
        self.filename = filename
        self.grammar = grammar or Grammar()
        self._closure_stack: t.List[str] = []  #: Used to construct nested closure names.
        self._closure_counter = 0  #: Used to assign a unique number to every closure.
        self._closures: t.Dict[str, Closure] = {}

    @contextlib.contextmanager
    def _lookahead(self) -> t.Iterator[t.Callable[[], None]]:
        """
        Context manager to save the current tokenizer and closure state and restore it on exit. This is
        useful for lookaheads, like :meth:`_test_dict`. If the returned callable is called, the
        tokenizer and closure state is not restored.
        """

        state = self.tokenizer.state
        closure_state = self._closure_counter, self._closures.copy(), self._closure_stack[:]
        do_restore = True

        def commit() -> None:
            nonlocal do_restore
            do_restore = False

        try:
            yield commit
        finally:
            if do_restore:
                self.tokenizer.state = state
                self._closure_counter, self._closures, self._closure_stack = closure_state

    def _syntax_error(self, msg: str, pos: t.Optional[Cursor] = None) -> SyntaxError:
        """Raise a syntax error on the current position of the tokenizer, or the specified *pos*."""

        pos = pos or self.tokenizer.current.pos
        text = self.tokenizer.scanner.getline(pos)
        return SyntaxError(msg, self.filename, pos.line, pos.column, text)

    @debug_trace
    def _consume_whitespace(self, newlines: t.Union[bool, ParseMode] = False, reset_to_indent: bool = True) -> str:
        """
        Consumes whitespace, indents, comments, and, if enabled, newlines until a different token is
        encountered. If *reset_to_indent* is enabled (default) then the tokenizer will be moved back
        to the indent token before that different token.
        """

        if isinstance(newlines, ParseMode):
            newlines = bool(newlines & ParseMode.GROUPED)

        token = ProxyToken(self.tokenizer)
        parts: t.List[str] = []
        state = token.save()
        while token.is_ignorable(newlines):
            parts.append(token.value)
            state = token.save()
            token.next()
        if reset_to_indent and state.token and state.token.type == Token.Indent:
            token.load(state)
            parts.pop()
        return "".join(parts)

    @debug_trace
    def _parse_closure(self) -> t.Optional[Closure]:
        """
        Attempts to parse a closure at the current position of the tokenizer. Closures can have the
        following syntactical variants:

        1. `() -> { stmts }`
        2. `arg -> { stmts }`
        3. `(arg1, arg2) -> { stmts }`
        4. `{ stmts }`

        Closures of the fourth form can conflict syntactically with Python set literals and will thus
        override the native syntactic feature. The fourth form also results in the #node.Closure
        parameter's list to be `None`.

        The first to third form may also exist without curly braces to define a single expression
        as the closure body (returning the value of that expression from the closure).

        1. `() -> expr`
        2. `arg -> expr`
        3. `(arg1, arg2) -> expr`

        Returns #None if no closure can be parsed at the current position of the tokenizer.
        """

        token = ProxyToken(self.tokenizer)
        pos = token.pos
        state = token.save()
        arglist = self._parse_closure_header()
        body: t.Optional[str] = None
        expr: t.Optional[str] = None
        closure_id = "".join(self._closure_stack) + f"_closure_{self._closure_counter + 1}"
        self._closure_stack.append(closure_id)

        if token.tv == (Token.Control, "{"):
            body = self._parse_closure_body()
        if body is None and arglist is not None:
            # We only parse an expression for the Closure body if an arglist was specified.
            expr = self._rewrite_expr(mode=ParseMode.DEFAULT)

        assert self._closure_stack.pop() == closure_id

        if not (body or expr):
            # NOTE(NiklasRosenstein): We could raise our own SyntaxError here if an arglist was provided
            #     as that is a strong indicator that a Closure expression or body should be provided,
            #     but we can also just leave the complaining to the Python parser.
            token.load(state)
            return None

        self._closure_counter += 1
        return Closure(closure_id, pos.line, pos.column, arglist, body, expr)

    @debug_trace
    def _parse_closure_body(self) -> t.Optional[str]:
        """
        Parses the body of a closure and returns it's code. Expects the tokenizer to point to the
        opening curly brace of the closure.
        """

        token = ProxyToken(self.tokenizer)
        assert token.tv == (Token.Control, "{"), token
        token.next()

        code = self._consume_whitespace(True)
        if "\n" in code:  # Multiline closure
            code += self._rewrite_stmt_block() + self._consume_whitespace(True, False)
        else:  # Singleline closure
            while token.type != Token.Newline and token.tv != (Token.Control, "}"):
                code += self._rewrite_stmt_singleline() + self._consume_whitespace(True, False)

        if token.tv != (Token.Control, "}"):
            raise self._syntax_error("expected closure closing brace")

        token.next()
        return code

    @debug_trace
    def _parse_closure_header(self) -> t.Optional[t.List[str]]:
        """
        Handles the possible formats for a closure header, f.e. a single argument name or an arglist
        followed by an arrow (`->`). Returns `None` if there can be no closure header extracted from
        the current position of the lexer.
        """

        token = ProxyToken(self.tokenizer)
        state = token.save()

        with token.set_skipped(Token.Whitespace):
            arglist: t.Optional[t.List[str]] = None
            if token.tv == (Token.Control, "("):
                arglist = self._parse_closure_arglist()
            elif token.type == Token.Name:
                arglist = [token.value]
                token.next()

            if arglist is None or token.tv != (Token.Control, "->"):
                # We may have found something that looks like an arglist, but isn't, or we found an
                # arglist but no following arrow, so we go back to where we started and let someone
                # else handle these tokens.
                token.load(state)
                return None

            token.next()
            return arglist

    @debug_trace
    def _parse_closure_arglist(self) -> t.Optional[t.List[str]]:
        """
        This method expects an open parenthesis as the current token and attempts to extract a list of
        argument names. Returns `None` if no argument list was actually extracted.
        """

        token = ProxyToken(self.tokenizer)
        assert token.tv == (Token.Control, "("), token

        state = token.save()
        with token.set_skipped({Token.Whitespace, Token.Comment, Token.Newline, Token.Indent}):
            token.next()

            arglist: t.List[str] = []
            is_delimited = True

            while token.tv != (Token.Control, ")"):

                if (
                    not is_delimited  # Token is not preceeded by an opening parentheses or comma.
                    or token.type != Token.Name
                ):  # We can only accept a name at this position.
                    token.load(state)
                    return None

                arglist.append(token.value)
                token.next()
                is_delimited = token.tv == (Token.Control, ",")
                if is_delimited:
                    token.next()

            assert token.tv == (Token.Control, ")"), token

            token.next()
            return arglist

    @debug_trace
    def _rewrite_expr(self, mode: ParseMode) -> str:
        """
        Consumes a Python expression and returns it's code. Does not parse over a comma.

        :param mode: The current parse mode that provides context about the level that is currently
          being parsed.
        """

        code = self._consume_whitespace(mode, False)
        code += self._rewrite_atom(mode)

        token = ProxyToken(self.tokenizer)
        while token:
            code += self._consume_whitespace(mode)

            if token.type == Token.Control and token.value in BINARY_OPERATORS:
                code += token.value
                token.next()
                code += self._rewrite_expr(mode)

            elif token.is_control("(["):
                code += self._consume_whitespace(True, False)
                code += self._rewrite_atom(
                    ParseMode.FUNCTION_CALL | ParseMode.GROUPED if token.value == "(" else ParseMode.DEFAULT
                )

            else:
                break

        return code

    @debug_trace
    def _find_current_line_indent(self) -> int:
        with self._lookahead():
            self.tokenizer.scanner.pos = self.tokenizer.scanner.pos.get_line_begin()
            token = self.tokenizer.next()
            assert token.type == Token.Indent
            return len(token.value)

    @debug_trace
    def _check_next_indent(self, min_indent: int) -> t.Union[int, None]:
        with self._lookahead():
            token = ProxyToken(self.tokenizer)
            assert token.type in (Token.Newline, Token.Indent), token
            whitespace = self._consume_whitespace(True, False).splitlines()
            indent = len(whitespace[-1])
            if indent < min_indent:
                return None
            return indent

    @debug_trace
    def _rewrite_items(self, mode: ParseMode) -> str:
        """
        Rewrites expressions separated by commas.
        """

        line_indent = self._find_current_line_indent()
        continuation_indent: t.Union[int, None] = None

        token = ProxyToken(self.tokenizer)
        code = ""
        upsert_comma = False
        upgraded_to_call_args = False
        do_break = False

        while True:
            code += self._consume_whitespace(mode)

            if token.type == Token.Newline and not (mode & ParseMode.GROUPED):
                break

            with self._lookahead() as commit:
                try:
                    code += self._rewrite_expr(mode=mode)
                except SyntaxError:
                    logger.debug("syntax error consumed while trying to parse expression", exc_info=sys.exc_info())
                    break
                commit()

            code += self._consume_whitespace(mode)

            if mode & ParseMode.CALL_ARGS and (
                token.is_control("=") or (self.grammar.colon_kwargs and token.is_control(":"))
            ):
                code += "="
                token.next()
                # TODO(NiklasRosenstein): This may be problematic in unparenthesised calls?
                code += self._rewrite_expr(mode=mode)

            if token.is_control(","):
                code += ","
                token.next()
                upsert_comma = False

            elif mode & ParseMode.CALL_ARGS and self.grammar.nocomma_args:
                code += ","
                upsert_comma = True

            else:
                do_break = True

            if token.type == Token.Newline:
                if mode & ParseMode.GROUPED:
                    pass

                elif continuation_indent is None:
                    continuation_indent = self._check_next_indent(line_indent + 1)
                    if continuation_indent is None:
                        break
                    if mode == ParseMode.DEFAULT and self.grammar.unparen_calls:
                        mode = ParseMode.CALL_ARGS
                        code += "("
                        upgraded_to_call_args = True

                elif self._check_next_indent(continuation_indent) is None:
                    break
                code += self._consume_whitespace(True)
                continue

            if do_break:
                break

        if upsert_comma:
            code = code.rstrip(",")
        if upgraded_to_call_args:
            code += ")"
        return code

    @debug_trace
    def _rewrite_atom(self, mode: ParseMode = ParseMode.DEFAULT) -> str:
        """
        Consumes a Python or BuildDSL language atom and returns it rewritten as pure Python code. If
        a closure is encountered, it will be replaced with a name reference and the closure itself will
        be stored in the #_closures mapping.
        """

        token = ProxyToken(self.tokenizer)

        if token.is_control("{") and self._test_dict():
            return self._rewrite_dict()

        code = ""
        closure = self._parse_closure()
        if closure:
            code += closure.id
            self._closures[closure.id] = closure

        elif token.is_control("([{"):
            assert not (mode & ParseMode.FUNCTION_CALL) or token.is_control(
                "("
            ), "ParseMode.FUNCTION_CALL requires current token be opening parenthesis"

            expected_close_token = {"(": ")", "[": "]", "{": "}"}[token.value]
            code += token.value
            token.next()
            code += self._consume_whitespace(True)
            if not token.is_control(expected_close_token):
                new_mode = ParseMode.CALL_ARGS if mode & ParseMode.FUNCTION_CALL else ParseMode.DEFAULT
                code += self._rewrite_items(new_mode | ParseMode.GROUPED) + self._consume_whitespace(mode, False)
            if not token.is_control(expected_close_token):
                raise self._syntax_error(f"expected {expected_close_token} but got {token}")

            code += expected_close_token
            token.next()

        elif mode & ParseMode.CALL_ARGS and (token.is_control(["*", "**"])):
            code += token.value
            token.next()
            code += self._rewrite_expr(mode=ParseMode.DEFAULT)

        elif token.type in (Token.Name, Token.Literal):
            code += token.value
            token.next()

        elif token.type == Token.Control and token.value in UNARY_OPERATORS:
            code += token.value
            token.next()
            code += self._rewrite_expr(mode=mode)
            return code

        else:
            raise self._syntax_error(f"not sure how to deal with {token} {mode}")

        return code

    @debug_trace
    def _test_dict(self) -> bool:
        """
        Tests if the code from the current opening curly brace looks like a dictionary definition.
        This does not match an empty dictionary, but only one with at least one key.
        """

        token = ProxyToken(self.tokenizer)
        assert token.is_control("{"), False

        with self._lookahead():
            token.next()
            self._consume_whitespace(True, False)
            try:
                self._rewrite_expr(mode=ParseMode.GROUPED)
                self._consume_whitespace(True, False)
                return token.is_control(":")
            except SyntaxError:
                return False

    @debug_trace
    def _rewrite_dict(self) -> str:
        token = ProxyToken(self.tokenizer)
        assert token.is_control("{"), token
        token.next()
        code = "{"

        while not token.is_control("}"):
            code += self._consume_whitespace(True, False)
            code += self._rewrite_expr(mode=ParseMode.GROUPED)
            code += self._consume_whitespace(True, False)
            if not token.is_control(":"):
                raise self._syntax_error("expected :")
            code += ":"
            token.next()
            code += self._consume_whitespace(True, False)
            code += self._rewrite_expr(mode=ParseMode.GROUPED)
            code += self._consume_whitespace(True, False)
            if not token.is_control(","):
                break
            code += ","
            token.next()
            code += self._consume_whitespace(True, False)

        if not token.is_control("}"):
            raise self._syntax_error("expected }")

        token.next()
        return code + "}"

    @debug_trace
    def _rewrite_stmt_singleline(self) -> str:
        token = ProxyToken(self.tokenizer)
        code = self._consume_whitespace(False)

        if token.type == Token.Name and token.value == "pass":
            token.next()
            return code + "pass" + self._consume_whitespace(True)

        elif token.type == Token.Name and token.value in ("assert", "return", "yield"):
            code += token.value
            is_yield = token.value == "yield"
            token.next()
            code += self._consume_whitespace(False)
            if is_yield and token.tv == (Token.Name, "from"):
                code += token.value
                token.next()
            code += self._rewrite_items(ParseMode.DEFAULT) + self._consume_whitespace(True)
            return code

        elif token.type == Token.Name and token.value in ("import", "from"):
            while token.type != Token.Newline and token.tv != (Token.Control, ";"):
                code += token.value
                token.next()
            code += token.value
            token.next()
            return code

        else:
            return code + self._rewrite_stmt_line_expr_or_assign()

    @debug_trace
    def _rewrite_stmt_line_expr_or_assign(self) -> str:
        token = ProxyToken(self.tokenizer)
        code = self._rewrite_items(ParseMode.DEFAULT)

        if not code:
            # TODO (@nrosenstein): Better error message. How to reproduce reaching this line:
            #
            #   print('hello, world'!)
            #
            # Note how the exclamation mark is outside the string literal.
            raise self._syntax_error("unable to parse statement")

        code += self._consume_whitespace(newlines=False)

        if token.type == Token.Control and token.value in ASSIGNMENT_OPERATORS:
            op = token.value
            token.next()
            code += op + self._consume_whitespace(newlines=False) + self._rewrite_items(ParseMode.DEFAULT)

        elif token and not token.is_ignorable(True) and not token.is_control(")]}:") and self.grammar.unparen_calls:
            if code[-1].isspace():
                code = code[:-1]
            # TODO(NiklasRosenstein): We may want to indicate here that we're parsing call arguments,
            #   but that the call is not parenthesised.
            code += "(" + self._rewrite_items(ParseMode.CALL_ARGS) + ")"

        # TODO (@nrosenstein): This is a nasty hack to figure out if the current line contains _just_ a name or
        #   a dotted name which, with unparenthesized calls enabled, should act as a call without arguments. Since
        #   we'll end up parsing code that was previously re-written for example in the case of closures, we could
        #   end up parsing a closure name (eg. just "_closure_1") if a closure was defined on its own, however that
        #   causing the closure to be called immediately seems a bit of an erratic behaviour so we want to catch it.
        elif (
            not (set(code) - set(string.ascii_letters + string.digits + "." + "_")) and self.grammar.unparen_calls
        ) and not code.startswith("_closure_"):
            code += "()"

        return code + self._consume_whitespace(True)

    @debug_trace
    def _test_local_def(self) -> t.Optional[str]:
        """
        Tests if the current `def` keyword introduces a local variable assignment, and if so,
        returns the code for the rewritten code for the entire assignment.
        """

        token = ProxyToken(self.tokenizer)
        assert token.tv == (Token.Name, self.grammar.local_keyword), token

        with self._lookahead() as commit:
            token.next()
            self._consume_whitespace(False)
            if token.type != Token.Name:
                return None
            code = self.grammar.local_prefix + token.value
            token.next()
            code += self._consume_whitespace(False)
            if not token.is_control("="):
                return None
            code += token.value
            token.next()
            code += self._rewrite_expr(ParseMode.DEFAULT)
            commit()
            return code

    @debug_trace
    def _rewrite_stmt(self, indentation: int) -> str:
        """
        Parses a line statement of Python code. Returns an empty string if the actual indendation of
        the code is lower than *indentation*. Handles parsing of Python block statements (such as if,
        try, etc.) recursively.
        """

        code = self._consume_whitespace(True)

        token = ProxyToken(self.tokenizer)
        assert token.type == Token.Indent, token
        if len(token.value) < indentation:
            return ""
        elif len(token.value) > indentation:
            raise self._syntax_error("unexpected indentation")

        code += token.value
        token.next()

        if self.grammar.local_def and token.tv == (Token.Name, self.grammar.local_keyword):
            defcode = self._test_local_def()
            if defcode:
                return code + defcode

        if token.type == Token.Name and token.value in PYTHON_BLOCK_KEYWORDS:
            # Parse to the next colon.
            # TODO(nrosenstein): If we want to support BuildDSL syntax in the expressions of block
            #   statements, we'll need to rewrite them on a more granular level.
            while token and token.tv not in ((Token.Newline, "\n"), (Token.Control, ":")):
                code += token.value
                token.next()
            if token.tv != (Token.Control, ":"):
                raise self._syntax_error(f"expected semicolon, found {token}")
            code += ":"
            token.next()

            return code + self._rewrite_stmt_block(indentation)

        if token.is_control("}"):
            return code

        else:
            return code + self._rewrite_stmt_singleline()

    @debug_trace
    def _rewrite_stmt_block(self, parent_indentation: t.Optional[int] = None) -> str:
        """
        Rewrites an entire statement block and returns it's rewritten code.
        """

        token = ProxyToken(self.tokenizer)
        code = self._consume_whitespace(True)
        if not token:
            return code
        assert token.type == Token.Indent, token
        if parent_indentation is not None and len(token.value) <= parent_indentation:
            raise self._syntax_error(f"expected indent > {parent_indentation}, found {token}")
        indentation = len(token.value)
        while token:
            stmt = self._rewrite_stmt(indentation)
            if not stmt:
                break
            code += stmt + self._consume_whitespace(True)
        return code

    @debug_trace
    def rewrite(self) -> RewriteResult:
        """
        Rewrite the code and return the #RewriteResult. This can be interpreted by the
        #builddsl.transpiler.ClosureRewriter to re-inject the code for closures.
        """

        return RewriteResult(self._rewrite_stmt_block(), self._closures)



""" A superset of the Python programming language with support for closures and multi-line lambdas. """

from ._execute import execute
from ._rewriter import Grammar, SyntaxError
from ._runtime import ChainContext, Closure, Context, MapContext, ObjectContext, UnboundClosure
from ._transpiler import TranspileOptions, transpile_to_ast, transpile_to_source

__version__ = "1.0.0"

__all__ = [
    "execute",
    "Grammar",
    "SyntaxError",
    "ChainContext",
    "Closure",
    "Context",
    "MapContext",
    "ObjectContext",
    "UnboundClosure",
    "TranspileOptions",
    "transpile_to_ast",
    "transpile_to_source",
]



"""
Helpers for added runtime capabilities of transpiled BuildDSL code, specifically around variable
name resolution when enabling #TranspileOptions.closure_target.
"""

import builtins
import enum
import sys
import types
import typing as t
from dataclasses import dataclass

import typing_extensions as te

from ._transpiler import TranspileOptions, transpile_to_ast

# import weakref


class NotSet(enum.Enum):
    Value = 1


undefined = NotSet.Value


class Context(te.Protocol):
    """
    A context implements the behaviour for dynamic name resolution in programs transpiled with
    BuildDSL (when dynamic name resolution is enabled). It behaves essentially as a mapping
    that is used to get/set/delete variables.
    """

    def __getitem__(self, key: str) -> t.Any:
        ...

    def __setitem__(self, key: str, value: t.Any) -> None:
        ...

    def __delitem__(self, key: str) -> None:
        ...

    def chain_with(self, other: "Context") -> "ChainContext":
        return ChainContext(self, other)


class ObjectContext(Context):
    """
    Proxies an object's members for get/set/delete operations of the dynamic name resolution.

    Raises a #RuntimeError if there is an attempted to set or delete an attribute that appears
    to be a method of the object.
    """

    def __init__(self, target: t.Any) -> None:
        self._target = target

    def _error(self, key: str) -> NameError:
        raise NameError(f"object of type {type(self._target).__name__} does not have an attribute {key!r}")

    def __getitem__(self, key: str) -> t.Any:
        value = getattr(self._target, key, undefined)
        if value is not undefined:
            return value
        raise self._error(key)

    def __setitem__(self, key: str, value: t.Any) -> None:
        current = getattr(self._target, key, undefined)
        if current is undefined:
            raise self._error(key)
        if isinstance(current, types.MethodType) and current.__self__ is current:
            raise RuntimeError(f"cannot overwrite method {type(self._target).__name__}.{key}()")
        setattr(self._target, key, value)

    def __delitem__(self, key: str) -> None:
        current = getattr(self._target, key, undefined)
        if current is undefined:
            raise self._error(key)
        if isinstance(current, types.MethodType) and current.__self__ is current:
            raise RuntimeError(f"cannot delete method {type(self._target).__name__}.{key}()")
        delattr(self._target, key)


class MapContext(Context):
    """
    Delegates dynamic name resolution to a mapping.
    """

    def __init__(self, target: t.MutableMapping[str, t.Any], description: str) -> None:
        self._target = target
        self._description = description

    def _error(self, key: str) -> NameError:
        raise NameError(f"{self._description} does not have an attribute {key!r}")

    def __getitem__(self, key: str) -> t.Any:
        if key in self._target:
            return self._target[key]
        raise self._error(key)

    def __setitem__(self, key: str, value: t.Any) -> None:
        if key in self._target:
            self._target[key] = value
        else:
            raise self._error(key)

    def __delitem__(self, key: str) -> None:
        if key in self._target:
            del self._target[key]
        else:
            raise self._error(key)


class ChainContext(Context):
    """
    Chain multiple #Context implementations.
    """

    def __init__(self, *contexts: Context) -> None:
        self._contexts = contexts

    def __getitem__(self, key: str) -> t.Any:
        for ctx in self._contexts:
            try:
                return ctx[key]
            except NameError:
                pass
        raise NameError(key)

    def __setitem__(self, key: str, value: t.Any) -> None:
        for ctx in self._contexts:
            try:
                ctx[key] = value
                return
            except NameError:
                pass
        raise NameError(key)

    def __delitem__(self, key: str) -> None:
        for ctx in self._contexts:
            try:
                del ctx[key]
                return
            except NameError:
                pass
        raise NameError(key)

    def chain_with(self, other: Context) -> "ChainContext":
        return ChainContext(*self._contexts, other)


class Closure(Context):
    r"""
    This class serves as a mapping to use for dynamic name lookup when transpiling BuildDSL code to Python.
    Several options in the #TranspileOptions need to be tweaked for this to work correctly as the closure
    hierarchy needs to be built up manually:

    * Set #TranspileOptions.closure_target to `__closure__`
    * Set #TranspileOptions.closure_def_prefix to `@__closure__.subclosure\n`
    * Set #TranspileOptions.closure_arglist_prefix to `__closure__,`

    You can initialize a #TranspileOptions object with these values using #init_options().

    When resolving names using #__getitem__(), #__setitem__() and #__delitem__(), the names will be looked
    up in the hierarchy of the closure itself. However do note that #__setitem__() and #__delitem__() cannot
    apply changes to the locals in a function. This is handled by proper rewriting rules in the #NameRewriter.
    """

    @staticmethod
    def init_options(options: TranspileOptions) -> None:
        options.closure_target = "__closure__"
        options.closure_def_prefix = "@__closure__.subclosure\n"
        options.closure_arglist_prefix = "__closure__,"

    @staticmethod
    def get_options() -> TranspileOptions:
        options = TranspileOptions()
        Closure.init_options(options)
        options.sync()
        return options

    def __init__(
        self,
        parent: t.Optional["Closure"] = None,
        frame: t.Optional[types.FrameType] = None,
        target: t.Any = None,
        context_factory: t.Callable[[t.Any], Context] = ObjectContext,
        target_context: t.Optional[Context] = None,
    ) -> None:
        """
        Args:
          parent: The parent context. Dynamic name resolution is delegated to the parent if the current
            context fails to provide it.
          frame: The Python frame in which the closure is defined.
          target: The target object that is wrapped by the closure.
          context_factory: A factory to create #Context#s for closures defined inside this closure.
          target_context: The context for this closure. If not given, it will be created using the
            *context_factory*.
        """

        self._parent = parent
        self._frame = frame  # weakref.ref(frame) if frame else None  # NOTE (@NiklasRosenstein): Cannot create weakref to frame  # noqa: E501
        self._target = target
        self._target_context = (
            (context_factory(target) if target is not None else None) if target_context is None else target_context
        )
        self._context_factory = context_factory

    def __repr__(self) -> str:
        return f"Closure(target={self._target!r})"

    @property
    def frame(self) -> t.Optional[types.FrameType]:
        return self._frame

    def subclosure(self, func: t.Callable[..., t.Any], frame: t.Optional[types.FrameType] = None) -> "UnboundClosure":
        if frame is None:
            frame = sys._getframe(1)

        closure = UnboundClosure(self, frame, func)
        del frame

        return closure

    def run_code(
        self,
        code: str,
        filename: str = "<string>",
        options: t.Optional[TranspileOptions] = None,
        scope: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> None:
        """
        Executes the given BuildDSL *code* with the #Closure as it's entry `__closure__` object.
        """

        if options:
            Closure.init_options(options)
        else:
            options = Closure.get_options()
        module = compile(transpile_to_ast(code, filename, options), filename, "exec")
        if scope is None:
            scope = {}
        assert options.closure_target
        scope[options.closure_target] = self
        exec(module, scope)

    def __getitem__(self, key: str) -> t.Any:
        frame = self.frame
        if frame and key in frame.f_locals:
            return frame.f_locals[key]
        if self._target_context is not None:
            try:
                return self._target_context[key]
            except NameError:
                pass
        if self._parent is not None:
            try:
                return self._parent[key]
            except NameError:
                pass
        if hasattr(builtins, key):
            return getattr(builtins, key)
        raise NameError(f"{key!r} in {self!r}")

    def __setitem__(self, key: str, value: t.Any) -> None:
        frame = self.frame
        if frame and key in frame.f_locals:
            raise RuntimeError("cannot set local variable through context, this should be handled by the transpiler")
        if self._target_context is not None:
            try:
                self._target_context[key] = value
                return
            except NameError:
                pass
        if self._parent is not None:
            try:
                self._parent[key] = value
                return
            except NameError:
                pass
        raise NameError(f"unclear where to set {key!r}")

    def __delitem__(self, key: str) -> None:
        frame = self.frame
        if frame and key in frame.f_locals:
            raise RuntimeError("cannot delete local variable through context, this should be handled by the transpiler")
        if self._target_context is not None:
            try:
                del self._target_context[key]
                return
            except NameError:
                pass
        if self._parent is not None:
            try:
                del self._parent[key]
                return
            except NameError:
                pass
        raise NameError(f"unclear where to delete {key!r}")

    @staticmethod
    def from_map(m: t.MutableMapping[str, t.Any]) -> "Closure":
        return Closure(None, None, m, ObjectContext, MapContext(m, "Closure.from_map()"))


@dataclass
class UnboundClosure:
    parent: Closure
    frame: types.FrameType
    func: t.Callable[..., t.Any]

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        __closure__ = Closure(self.parent, self.frame, args[0] if args else None, self.parent._context_factory)
        return self.func(__closure__, *args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['builddsl']

package_data = \
{'': ['*']}

install_requires = \
['nr-io-lexer>=1.0.0,<2.0.0', 'typing-extensions>=3.0.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.6,<0.7']}

setup_kwargs = {
    'name': 'builddsl',
    'version': '1.0.0',
    'description': 'A superset of the Python programming language with support for closures and multi-line lambdas.',
    'long_description': '# builddsl\n\nA superset of the Python programming language with support for closures and multi-line lambdas.\n\nBuildDSL is an "almost superset" of Python 3; adding a lot of syntactical features that make it more\nconvenient to describe build configurations at the cost of some other syntax features of the native Python\nlanguage (like set literals).\n\n## Installation\n\n    $ pip install builddsl\n\nThe `builddsl` package requires at least Python 3.8.\n\n## Projects using BuildDSL\n\n* [Novella](https://niklasrosenstein.github.io/novella/)\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import ast
import contextlib
import dataclasses
import typing as t

T_AST = t.TypeVar("T_AST", bound=ast.AST)


@dataclasses.dataclass
class DynamicLookupRewriter(ast.NodeTransformer):
    """Rewrites names in  be accessed through a lookup object.

    Imports and variables beginning with #ignore_prefix are treated as pure builtins for
    their scope (i.e. they will not be substituted with a dynamic lookup).
    """

    # TODO(NiklasRosenstein): Handle more nodes that define local variables and := operator.

    #: The variable name of the target object that name resolution should occur through.
    #: All names in the AST, with a few exceptions, will be replaced by a getitem/setitem/delitem
    #: expression on the variable name defined here.
    lookup_target: str

    #: Names to not replace. The #lookup_target does not need to be added here explicitly.
    pure_builtins: t.Collection[str] = dataclasses.field(default_factory=frozenset)

    #: A prefix to compare variable names for which, if it matches, they will not be replaced
    #: with a dynamic lookup, but instead the prefix will be trimmed.
    ignore_prefix: "str | None" = None

    def __post_init__(self) -> None:
        self._locals: t.List[t.Set[str]] = [set()]

    def _add_to_locals(self, varnames: t.Set[str]) -> None:
        assert self._locals, "no locals in current scope"
        self._locals[-1].update(varnames)

    @contextlib.contextmanager
    def _with_locals(self, varnames: t.Set[str]) -> t.Iterator[None]:
        self._locals.append(varnames)
        try:
            yield
        finally:
            self._locals.pop()

    @contextlib.contextmanager
    def _with_locals_from_target(self, target: ast.expr) -> t.Iterator[None]:
        names: t.Set[str] = set()
        if isinstance(target, ast.Name):
            names.add(target.id)
        elif isinstance(target, (ast.List, ast.Tuple)):
            names.update(n.id for n in target.elts)  # type: ignore  # TODO (@NiklasRosenstein)
        else:
            raise TypeError(f"expected Name/List/Tuple, got {type(target).__name__}")
        with self._with_locals(names):
            yield

    def _has_local(self, varname: str) -> bool:
        if self._locals:
            return varname in self._locals[-1]
        return False

    def _has_nonlocal(self, varname: str) -> bool:
        for locals in self._locals:
            if varname in locals:
                return True
        return varname == self.lookup_target or varname in self.pure_builtins

    def visit_Name(self, node: ast.Name) -> ast.AST:
        if self._has_nonlocal(node.id):
            return node
        return ast.Subscript(
            value=ast.Name(id=self.lookup_target, ctx=ast.Load()),
            slice=ast.Index(value=ast.Constant(value=node.id)),
            ctx=node.ctx,
        )

    def visit_Assign(self, assign: ast.Assign) -> ast.AST:
        if len(assign.targets) == 1 and isinstance(assign.targets[0], ast.Name):
            name = assign.targets[0]
            if self.ignore_prefix is not None and name.id.startswith(self.ignore_prefix):
                name.id = name.id[len(self.ignore_prefix) :]
                self._add_to_locals({name.id})
        return self.generic_visit(assign)

    def visit_For(self, node: ast.For) -> ast.AST:
        with self._with_locals_from_target(node.target):
            return self.generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.AST:
        self._add_to_locals({node.name})
        names: t.Set[str] = set()
        for arg in node.args.args:
            names.add(arg.arg)
        for arg in node.args.kwonlyargs:
            names.add(arg.arg)
        if node.args.vararg:
            names.add(node.args.vararg.arg)
        if node.args.kwarg:
            names.add(node.args.kwarg.arg)
        with self._with_locals(names):
            return self.generic_visit(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.AST:
        self._add_to_locals({node.name})
        return self.generic_visit(node)

    def visit_Import(self, node: ast.Import) -> ast.AST:
        names: t.Set[str] = set()
        for name in node.names:
            if name.asname:
                names.add(name.asname)
            elif "." in name.name:
                names.add(name.name.rpartition(".")[0])
            else:
                names.add(name.name)
        self._add_to_locals(names)
        return self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.AST:
        self.visit_Import(ast.Import(names=node.names))  # Dispatch name detection
        return self.generic_visit(node)
